@@ -1,7 +1,11 @@
 import java.io.File;
 import java.util.Arrays;
+import java.util.function.BiConsumer;
+import java.util.function.Function;
 import java.util.LinkedList;
 import java.util.List;
+import java.util.stream.Collectors;
+import java.util.stream.Stream;
 
 import java.math.BigInteger;
 
@@ -78,7 +82,7 @@ public final class BindingsTests {
 
         // Prepare the working variables
         AnalysisUnit unit;
-        List<Diagnostic> diags;
+        Diagnostic[] diags;
 
         try(
             AnalysisContext context = AnalysisContext.create()
@@ -114,6 +118,20 @@ public final class BindingsTests {
         ) {
             // Get the unit for the foo.txt file
             AnalysisUnit unit = context.getUnitFromFile("foo.txt");
+            AnalysisUnit other = context.getUnitFromFile("foo.txt");
+            AnalysisUnit notSame =
+                context.getUnitFromBuffer("var identifier", "bar.txt");
+            System.out.println(
+                unit.toString() + " != " + other + " = " + (unit != other)
+            );
+            System.out.println(
+                unit.toString() + ".equals(" + other + ") = " +
+                unit.equals(other)
+            );
+            System.out.println(
+                unit.toString() + ".equals(" + notSame +
+                ") = " + unit.equals(notSame)
+            );
             System.out.println(
                 "Unit for the file " +
                 new File(unit.getFileName()).getName()
@@ -168,7 +186,7 @@ public final class BindingsTests {
             System.out.println("Text range with a NO_TOKEN = " +
                 Token.textRange(first, none));
 
-            // Test the token equivalence
+            // Test the token equivalence and equality
             AnalysisUnit eqUnit = context.getUnitFromBuffer(
                 "null identifier example identifier example",
                 "foo.txt"
@@ -189,13 +207,26 @@ public final class BindingsTests {
                 }
                 current = current.next();
             }
+
+            Token left = eqUnit.getFirstToken();
+            Token right = eqUnit.getFirstToken();
+            while(!(left.isNone() || right.isNone())) {
+                System.out.println(
+                    left.toString() + " != " + right + " = " + (left != right)
+                );
+                System.out.println(
+                    left.toString() + ".equals(" + right + ") = " +
+                    left.equals(right)
+                );
+                left = left.next();
+                right = right.next();
+            }
         }
 
         // Display the footer
         footer("Tokens");
     }
 
-
     /**
      * Test the node manipulation
      */
@@ -211,18 +242,18 @@ public final class BindingsTests {
 
             System.out.println("Unit = " + unit.toString());
             System.out.println("Unit root = " + root.toString());
-            System.out.println("Unit root kind = " + root.getKindName());
+            System.out.println("Unit root kind = " + root.getClassName());
             System.out.println(
                 "Unit root children count = " + root.getChildrenCount()
             );
             System.out.println(
-                "Unit root children = " + root.children().toString()
+                "Unit root children = " + Arrays.toString(root.children())
             );
             System.out.println(
-                "Unit root AST dump = " + root.dumpAST()
+                "Unit root tree dump = " + root.dumpTree()
             );
             System.out.println(
-                "Unit root is a list node = " + root.isListType()
+                "Unit root is a list node = " + root.isListNode()
             );
             System.out.println(
                 "Unit root is a token node = " + root.isTokenNode()
@@ -230,8 +261,18 @@ public final class BindingsTests {
             System.out.println("Unit root image = " + root.getImage());
             System.out.println("Unit root text = " + root.getText());
             Sequence sequence = (Sequence) root;
-            FooNodeArray items = sequence.pAllItems();
-            System.out.println("Root \"p_all_items\" = " + items.toString());
+            System.out.println("Iterating on root sequence:");
+            for (FooNode node : sequence) {
+                System.out.println("  - " + node.toString());
+            }
+            FooNode[] items = sequence.pAllItems();
+            System.out.println(
+                "Root \"p_all_items\" = " + Arrays.toString(items)
+            );
+            Example[] examples = sequence.pExampleItems();
+            System.out.println(
+                "Root \"p_example_items\" = " + Arrays.toString(examples)
+            );
             Var var = (Var) root.getChild(2);
             System.out.println("Var (3rd child) = " + var.toString());
             System.out.println("Var image = " + var.getImage());
@@ -239,14 +280,43 @@ public final class BindingsTests {
                 "Var fields = " + Arrays.toString(var.getFieldNames())
             );
             Sequence arg = var.fArg();
-            FooNodeArray argItems = arg.pAllItems();
+            FooNode[] argItems = arg.pAllItems();
             System.out.println("Var arg = " + arg.toString());
-            System.out.println("Var content = " + argItems.toString());
+            System.out.println("Var content = " + Arrays.toString(argItems));
             Var var2 = (Var) root.getChild(2);
             System.out.println("Node equality = " + var.equals(var2));
             System.out.println(
                 "Node hash equality = " + (var.hashCode() == var2.hashCode())
             );
+            System.out.println(
+                "Node 'p_all_items' member reference equality = " +
+                (root.getFieldDescription("p_all_items").memberRef ==
+                 MemberReference.FOO_SEQUENCE_P_ALL_ITEMS)
+            );
+
+            // Test node equality and hash codes
+            FooNode e1 = root.pWithMd(true, false);
+            FooNode e2 = root.pWithMd(true, true);
+            FooNode e3 = root.pWithMd(false, true);
+            System.out.println("e1.equals(e2) = " + e1.equals(e2));
+            System.out.println("e1.equals(e3) = " + e1.equals(e3));
+            System.out.println(
+                "hash(e1) == hash(e2) = " + (e1.hashCode() == e2.hashCode())
+            );
+            System.out.println(
+                "hash(e1) == hash(e3) = " + (e1.hashCode() == e3.hashCode())
+            );
+
+            // Test empty struct nullity
+            EmptyStruct emptyStruct = root.pNewEmptyStruct();
+            System.out.println(
+                "empty struct is_struct_null() = " +
+                root.pIsStructNull(emptyStruct)
+            );
+            System.out.println(
+                "null empty struct is_struct_null() = " +
+                root.pIsStructNull(EmptyStruct.NONE)
+            );
         }
 
         // Display the footer
@@ -306,7 +376,7 @@ public final class BindingsTests {
         ) {
             AnalysisUnit unit = context.getUnitFromFile("foo.txt");
             Sequence root = (Sequence) unit.getRoot();
-            FooNodeArray items = root.pAllItems();
+            FooNode[] items = root.pAllItems();
 
             // Visit all the items and display their siblings
             for(FooNode item : items) {
@@ -338,21 +408,113 @@ public final class BindingsTests {
             Sequence root = (Sequence) unit.getRoot();
 
             // Visit the sequence
-            FooNodeArray items = root.pAllItems();
-            System.out.println("Sequence size = " + items.size());
+            FooNode[] items = root.pAllItems();
+            System.out.println("Sequence size = " + items.length);
             System.out.println("Sequence content :");
-            for(int i = 0 ; i < items.size() ; i++) {
-                FooNode item = items.get(i);
+            for(int i = 0 ; i < items.length ; i++) {
+                FooNode item = items[i];
                 System.out.println("  " + item.toString());
             }
-            FooNodeArray eqItems = root.pAllItems();
+            FooNode[] eqItems = root.pAllItems();
             System.out.println("Array equality = " + items.equals(eqItems));
+
+            // Test an array of primitive type (integer)
+            int[] intArray = root.pGetIntArray();
+            System.out.println("Integer array size = " + intArray.length);
+            System.out.println(
+                "Integer array size (from property) = " +
+                root.pArrayLen(intArray)
+            );
+            System.out.println("Integer array content :");
+            for (int i = 0; i < intArray.length; i++) {
+                System.out.println("  " + intArray[i]);
+            }
+
+            // Test a property error raised by an array typed property
+            try {
+                FooNode[] invalid = root.pArrayPropError();
+                System.out.println("No exception raised by 'array_prop_error'");
+            } catch (LangkitException exc) {
+                System.out.println(
+                    "Expected exception: " + exc.kind + " - " +
+                    exc.getMessage() + (
+                        exc.nativeStackTrace() != "" ?
+                            " (Stack trace available)" :
+                            " (NO STACK TRACE)"
+                    )
+                );
+            }
+
         }
 
         // Display the footer
         footer("Arrays");
     }
 
+    private static void testIterators() {
+        header("Iterators");
+
+        try (
+            AnalysisContext context = AnalysisContext.create();
+        ) {
+            AnalysisUnit unit = context.getUnitFromBuffer(
+                "my_ident",
+                "foo.txt"
+            );
+            Sequence root = (Sequence) unit.getRoot();
+
+            // Get and display an integer iterator
+            System.out.println("--- Integer iterator");
+            try (
+                IntegerIterator intIterator = root.pIterInt()
+            ) {
+                while (intIterator.hasNext()) {
+                    int i = intIterator.next();
+                    System.out.println("  " + i);
+                }
+            }
+
+            // Get and display an entity iterator
+            System.out.println("--- Entity iterator");
+            try (
+                FooNodeIterator entityIterator = root.pIterEntity()
+            ) {
+                while (entityIterator.hasNext()) {
+                    FooNode i = entityIterator.next();
+                    System.out.println("  " + i);
+                }
+            }
+
+            // Get and display a struct iterator (ensure handing of null values)
+            System.out.println("--- Struct iterator");
+            try (
+                SomeStructIterator structIterator = root.pIterStruct()
+            ) {
+                while (structIterator.hasNext()) {
+                    SomeStruct i = structIterator.next();
+                    System.out.println(
+                        "  SomeStruct(examples=" +
+                        Arrays.toString(i.examples) +
+                        ")"
+                    );
+                }
+            }
+
+            // Check passing an iterator as property parameter
+            System.out.println("--- Iterator as property parameter");
+            try (
+                IntegerIterator iterator = root.pIterInt();
+                IntegerIterator same = root.pIterIntId(iterator)
+            ) {
+                System.out.println(
+                    "Identity iterator is the same = " + iterator.equals(same)
+                );
+            }
+        }
+
+        footer("Iterators");
+    }
+
     private static void testSymbols() {
         // Display the header
         header("Symbols");
@@ -415,18 +577,18 @@ public final class BindingsTests {
             FooNode root = unit.getRoot();
 
             // Test the character manipulation
-            LibfoolangField getAField = root.getFieldDescription("p_get_a");
-            LibfoolangField getEacuteField =
+            Reflection.Field getAField = root.getFieldDescription("p_get_a");
+            Reflection.Field getEacuteField =
                 root.getFieldDescription("p_get_eacute");
             Char c1 = root.pGetA(
                 (Char) (
-                    (ParamWithDefaultValue) getAField.params.get(0)
-                ).defaultValue
+                    (Reflection.Param) getAField.params.get(0)
+                ).defaultValue.get()
             );
             Char c2 = root.pGetEacute(
                 (Char) (
-                    (ParamWithDefaultValue) getEacuteField.params.get(0)
-                ).defaultValue
+                    (Reflection.Param) getEacuteField.params.get(0)
+                ).defaultValue.get()
             );
             Char cIdent = root.pIdentity(Char.create('\u00e9'));
             System.out.println("The 'a' char = " + c1.toString());
@@ -528,18 +690,162 @@ public final class BindingsTests {
             AnalysisUnit unit = context.getUnitFromFile("foo.txt");
             FooNode root = unit.getRoot();
 
+            // Test a struct with an entity
             FooNodeStruct falseNodeStruct = root.pMe(false);
             FooNodeStruct trueNodeStruct = root.pMe(true);
             FooNode falseNode = root.pGetNode(falseNodeStruct);
             FooNode trueNode = root.pGetNode(trueNodeStruct);
             System.out.println("me false = " + falseNode);
             System.out.println("me true = " + trueNode);
+            System.out.println(
+                "falseNode.equals(falseNodeStruct.node) = " +
+                falseNode.equals(falseNodeStruct.node)
+            );
+
+            // Test a struct with a bare node and verify that it has been
+            // wrapped.
+            BareNodeStruct bareNodeStruct = root.pMyNode();
+            FooNode bareNode = bareNodeStruct.bareNode;
+            System.out.println("root.pMyNode().bareNode = " + bareNode);
+
+            // Test passing this structure as a property parameter
+            Char aChar = bareNodeStruct.aChar;
+            Char otherChar = root.pGetChar(bareNodeStruct);
+            System.out.println(
+                "aChar.equals(otherChar) = " + aChar.equals(otherChar)
+            );
         }
 
         // Display the footer
         footer("Struct");
     }
 
+    private static void testEventHandlers() {
+        header("Event handlers");
+
+        // Local function to test event handlers
+        BiConsumer<
+            EventHandler.UnitRequestedCallback,
+            EventHandler.UnitParsedCallback
+        > testHandler = (unitRequestedCallback, unitParsedCallback) -> {
+            // Create the context with the event handler and create units
+            try(
+                EventHandler eventHandler = EventHandler.create(
+                    unitRequestedCallback,
+                    unitParsedCallback
+                );
+                AnalysisContext context = AnalysisContext.create(
+                    null,
+                    null,
+                    null,
+                    eventHandler,
+                    true,
+                    8
+                );
+            ) {
+                // Parse the unit twice to test the "reparsed" callback param
+                AnalysisUnit unit = context.getUnitFromBuffer(
+                    "example",
+                    "example"
+                );
+                unit = context.getUnitFromBuffer("example example", "example");
+
+                // Call the property to trigger "unit requested" callback
+                FooNode root = unit.getRoot();
+                root.pTriggerUnitRequested(
+                    Symbol.create("foo_a"),
+                    false,
+                    false
+                );
+                root.pTriggerUnitRequested(
+                    Symbol.create("foo_b"),
+                    true,
+                    false
+                );
+                root.pTriggerUnitRequested(
+                    Symbol.create("foo_c"),
+                    true,
+                    true
+                );
+                root.pTriggerUnitRequested(
+                    Symbol.create("foo_d"),
+                    false,
+                    true
+                );
+            }
+        };
+
+        // Create the callback functions
+        EventHandler.UnitRequestedCallback unitRequestedCallback = (
+            AnalysisContext context,
+            String name,
+            AnalysisUnit from,
+            boolean found,
+            boolean isNotFoundError
+        ) -> {
+            System.out.println("--- Unit requested callback");
+            System.out.println("name: " + name);
+            System.out.println("from: " + from);
+            System.out.println("found: " + found);
+            System.out.println("is_not_found_error: " + isNotFoundError);
+            System.out.println();
+        };
+        EventHandler.UnitParsedCallback unitParsedCallback = (
+            AnalysisContext context,
+            AnalysisUnit unit,
+            boolean reparsed
+        ) -> {
+            System.out.println("--- Unit parsed callback");
+            System.out.println("unit: " + unit);
+            System.out.println("reparsed: " + reparsed);
+            System.out.println();
+        };
+
+        System.out.println("=== Non null callbacks ===");
+        testHandler.accept(
+            unitRequestedCallback,
+            unitParsedCallback
+        );
+        System.out.println("=== Null callbacks ===");
+        testHandler.accept(null, null);
+
+        System.out.println("=== Unclosed event handler ===");
+        EventHandler dontClose = EventHandler.create(
+            unitRequestedCallback,
+            unitParsedCallback
+        );
+
+        footer("Event handlers");
+    }
+
+    private static class FooVisitor extends DefaultVisitor<Integer> {
+
+        protected Function<FooNode, Integer> createDefaultBehavior() {
+            return (node) -> {
+                System.out.println("Visiting " + node.getImage());
+                Integer count = 1;
+                for (var c : node.children()) {
+                    if (!c.isNone()) count += c.accept(this);
+                }
+                return count;
+            };
+        }
+    }
+
+    private static void testDefaultVisitor() {
+        header("Default visitor");
+        FooVisitor visitor = new FooVisitor();
+
+        try(
+            AnalysisContext context = AnalysisContext.create()
+        ) {
+            AnalysisUnit unit = context.getUnitFromFile("foo.txt");
+            Integer count = unit.getRoot().accept(visitor);
+            System.out.println("Visited " + String.valueOf(count) + " nodes");
+        }
+
+        footer("Default visitor");
+    }
     /**
      * Run the Java tests one by one
      *
@@ -556,11 +862,14 @@ public final class BindingsTests {
         testParent();
         testSiblings();
         testArrays();
+        testIterators();
         testSymbols();
         testCharacter();
         testString();
         testBigInt();
         testStruct();
+        testEventHandlers();
+        testDefaultVisitor();
         System.out.println("===== End of the Java tests =====");
     }
 
